import { AbortController } from 'abort-controller';
import { fromByteArray, toByteArray } from 'base64-js';
import { NativeEventEmitter, NativeModules, Platform } from 'react-native';
import { GrpcError } from './GrpcError';
import { GrpcCall, ServerOutputStream } from './GrpcCall';
const LINKING_ERROR = `The package '@matejdr/react-native-grpc-client' doesn't seem to be linked. Make sure: \n\n` + Platform.select({
  ios: "- You have run 'pod install'\n",
  default: ''
}) + '- You rebuilt the app after installing the package\n' + '- You are not using Expo managed workflow\n';
const GrpcBridge = NativeModules.GrpcBridge ? NativeModules.GrpcBridge : new Proxy({}, {
  get() {
    throw new Error(LINKING_ERROR);
  }

});
const Emitter = new NativeEventEmitter(NativeModules.GrpcBridge);

function createDeferred(signal) {
  let completed = false;
  const deferred = {};
  deferred.promise = new Promise((resolve, reject) => {
    deferred.resolve = value => {
      completed = true;
      resolve(value);
    };

    deferred.reject = reason => {
      completed = true;
      reject(reason);
    };
  });
  signal.addEventListener('abort', () => {
    if (!completed) {
      deferred.reject('aborted');
    }
  });
  return deferred;
}

let idCtr = 1;
const deferredMap = {};

function handleGrpcEvent(event) {
  var _deferred$headers, _deferred$data, _deferred$response, _deferred$trailers, _deferred$data2, _deferred$headers2, _deferred$trailers2, _deferred$response2, _deferred$data3;

  const deferred = deferredMap[event.id];

  if (deferred) {
    switch (event.type) {
      case 'headers':
        (_deferred$headers = deferred.headers) === null || _deferred$headers === void 0 ? void 0 : _deferred$headers.resolve(event.payload);
        break;

      case 'response':
        const data = toByteArray(event.payload);
        (_deferred$data = deferred.data) === null || _deferred$data === void 0 ? void 0 : _deferred$data.notifyData(data);
        (_deferred$response = deferred.response) === null || _deferred$response === void 0 ? void 0 : _deferred$response.resolve(data);
        break;

      case 'trailers':
        (_deferred$trailers = deferred.trailers) === null || _deferred$trailers === void 0 ? void 0 : _deferred$trailers.resolve(event.payload);
        (_deferred$data2 = deferred.data) === null || _deferred$data2 === void 0 ? void 0 : _deferred$data2.notifyComplete();
        delete deferredMap[event.id];
        break;

      case 'error':
        const error = new GrpcError(event.error, event.code);
        (_deferred$headers2 = deferred.headers) === null || _deferred$headers2 === void 0 ? void 0 : _deferred$headers2.reject(error);
        (_deferred$trailers2 = deferred.trailers) === null || _deferred$trailers2 === void 0 ? void 0 : _deferred$trailers2.reject(error);
        (_deferred$response2 = deferred.response) === null || _deferred$response2 === void 0 ? void 0 : _deferred$response2.reject(error);
        (_deferred$data3 = deferred.data) === null || _deferred$data3 === void 0 ? void 0 : _deferred$data3.noitfyError(error);
        delete deferredMap[event.id];
        break;
    }
  }
}

function getId() {
  return idCtr++;
}

export class GrpcClient {
  constructor() {
    Emitter.addListener('grpc-call', handleGrpcEvent);
  }

  destroy() {
    Emitter.removeAllListeners('grpc-call');
  }

  startCall(config, method, requestHeaders, methodType) {
    const id = getId();
    const abort = new AbortController();
    abort.signal.addEventListener('abort', () => {
      GrpcBridge.cancelGrpcCall(id);
    });
    const headers = createDeferred(abort.signal);
    const trailers = createDeferred(abort.signal);
    const stream = new ServerOutputStream();
    deferredMap[id] = {
      headers,
      trailers,
      data: stream
    };
    GrpcBridge.startCall(id, config, method, requestHeaders || {}, methodType);
    const call = new GrpcCall(method, requestHeaders || {}, headers.promise, stream, trailers.promise, abort, data => {
      const requestData = fromByteArray(data);
      const obj = {
        data: requestData
      };
      return GrpcBridge.sendMessage(id, obj);
    }, () => {
      return GrpcBridge.finishClientStreaming(id);
    });
    call.then(result => result, () => abort.abort());
    return call;
  }

}
export const GrpcClientImpl = new GrpcClient();
export { GrpcBridge };
//# sourceMappingURL=GrpcClient.js.map